#!/usr/bin/perl -w
#
#  Convert h.def into h.inc
#

use strict;

my %typedefs;
my $where = 'top';
my $where_name;
my $seen_ret = 0;
my @seen_argl;
my $seen_ret_type;
my @final;
my %protos;

while (<>) {
    s/\#.*//;
    s/\s+$//;
    if (/^$/) { next }
    if (/^type\s+(\S+):/) {
        # type definition
        finishFunction();
        $where = 'type';
        $where_name = $1;
    } elsif (/^([-a-z0-9]+)\s*\((.*)\)/) {
        # function definition
        my $name = $1;
        my $args = $2;
        my $cname = "F" . join "", map{ucfirst} split/-/, $name;
        my @argl = split /,\s*/, $args;
        my $pname = "p";
        my $pcount = 0;
        if (@argl) {
            while (grep /^$pname\s*:/, @argl) {
                $pname = "p" . ++$pcount;
            }
        } else {
            $pname = "";
        }
        finishFunction();
        $where = 'fun';
        $where_name = $name;
        $seen_ret = 0;
        @seen_argl = @argl;
        if (/\)\s*:\s*(\S+)/) {
            $seen_ret_type = $1
        } else {
            $seen_ret_type = '';
            if ($name !~ /^put-/) { print STDERR "WARNING: $name has no return type\n" }
        }
        if ($seen_ret_type eq 'void') { $seen_ret_type = '' }
        print "// ($name ", join(' ', @argl), ")\n";
        print "static LPtr\n";
        print "$cname (LPtr $pname)\n";
        print "{\n";
        my @val;
        my $proto;
        foreach (@argl) {
            /^(\S+)\s*:\s*(\S+)/ or die "invalid parameter: $_";
            my ($p, $t) = ($1, $2);
            if ($t =~ s/<(.*?)\.\.\s*(.*?)>//) {
                push @val, "    if ($p < $1 || $p > $2)\n";
                push @val, "        throw LLispException(\"range-error\", \"parameter $p out of range $1..$2\");\n";
            }
            print "    ", fetchType($t,'declare',$t), " ", $p, " = ",
              fetchType($t,'function',''), "($pname);\n";
            $proto .= "$t ";
        }
        print @val;
        my $argc = @argl;
        push @final, "LSymbol::define(\"$name\", new LBuiltin($cname, $argc, $argc))";
        if ($seen_ret_type eq '') { $proto .= "-> void"; } else { $proto .= "-> $seen_ret_type" }
        $protos{$name} = $proto;
    } elsif (/^\s+(.*)/) {
        my $line = $1;
        if ($where eq 'type') {
            if ($line =~ /^(\S+)\s+(.*)/) {
                $typedefs{$where_name}{$1} = $2;
            } else {
                die "syntax error";
            }
        } elsif ($where eq 'fun') {
            if ($line =~ /^call\s+(.*)/) {
                my $fun = "$1(";
                my @call_args;
                foreach (@seen_argl) {
                    /^(\S+)\s*:\s*(\S+)/ or die "invalid parameter: $_";
                    my ($p, $t) = ($1, $2);
                    $t =~ s/<.*//;
                    push @call_args, argSubst(fetchType($t,'call','*'),$p);
                }
                $fun .= join ', ', @call_args;
                $fun .= ')';
                if ($seen_ret_type ne '') {
                    print "    return ", argSubst(fetchType($seen_ret_type,'return','*'),$fun), ";\n";
                    $seen_ret = 1;
                } else {
                    print "    ", $fun, ";\n";
                }
            } elsif ($line =~ /^\|\s?(.*)/) {
                print "    $1\n";
            } else {
                print "    ", $line, ";\n";
            }
            $seen_ret = 1 if /\breturn\b/;
        } else {
            die "syntax error";
        }
    }
}
finishFunction();

print "static void\n";
print "initGeneratedFunctions()\n";
print "{\n";
foreach (@final) {
    print "    $_;\n";
}
buildSetfs();
print "}\n\n";

sub finishFunction {
    if ($where eq 'fun') {
        if (!$seen_ret) { print "    return LPtr();\n"; }
        print "}\n\n";
    }
    $where = 'top';
}

sub fetchType {
    my ($type,$what,$default) = @_;
    if (!exists $typedefs{$type}) { die "unknown type '$type'" }
    if (!exists $typedefs{$type}{$what}) { $default }
    else { $typedefs{$type}{$what} };
}

sub argSubst {
    my ($pat,$fill) = @_;
    $pat =~ s/\*/$fill/g;
    return $pat;
}

sub buildSetfs {
    foreach (sort keys %protos) {
        if (/^put-(.*)/) {
            my $base = $1;
            if (!exists $protos{$base}) {
                print STDERR "WARNING: there is a function '$_', but no '$base'\n";
            } else {
                # my prototype is foo1 foo2 foo3 -> void
                # other must be   foo1 foo2 -> foo3
                my @mproto = split /\s+/, $protos{$_};
                my @oproto = split /\s+/, $protos{$base};
                if ($mproto[-1] ne 'void') {
                    print STDERR "WARNING: '$_' does not return void\n";
                }
                pop @mproto;
                pop @mproto; # "->"
                if ($mproto[-1] ne $oproto[-1]) {
                    print STDERR "WARNING: '$_' last arg does not match '$base' return type\n";
                } else {
                    my $eq = 1;
                    for (my $i = 0; $i < @mproto-1; ++$i) {
                        if ($mproto[$i] ne $oproto[$i]) { $eq = 0 }
                    }
                    if ($eq) {
                        # ok
                        print "    LSymbol::defineSetf(\"$base\", \"$_\");\n";
                    } else {
                        print STDERR "WARNING: '$_' args do not match '$base'\n";
                    }
                }
            }
        }
    }
}
